define([    'shader!vertShader.vert',
            'shader!compShader.frag',
            'shader!initShader.frag',
            'Abubu/Abubu'
            ],
function(   vertShader,
            compShader,
            initShader,
            Abubu
            ){
"use strict" ;

/*========================================================================
 * Global Parameters
 *========================================================================
 */
var log = console.log ;
var env = {} ;
window.env = env ;

/*========================================================================
 * Initialization of the GPU and Container
 *========================================================================
 */
function loadWebGL()
{
    env.size  = 512 ;
    env.width = env.size ;
    env.height = env.size ;

    env.dt = 0.05 ;

/*------------------------------------------------------------------------
 * Access and setup canvas 
 *------------------------------------------------------------------------
 */
    var canvas_1 = document.getElementById("canvas_1") ;
    canvas_1.width = env.width ;
    canvas_1.height = env.height ;
 
    env.ftrgt = new Abubu.Float32Texture(env.width, env.height ) ;
    env.strgt = new Abubu.Float32Texture(env.width, env.height ) ;

    env.init = new Abubu.Solver({
            fragmentShader : initShader.value,
            vertexShader : vertShader.value ,
            targets : { 
                outTrgt1 : { location : 0, target : env.ftrgt } ,
                outTrgt2 : { location : 1, target : env.strgt } ,
            } 
        } ) ;

    env.aa= 0.1 ;
    env.bb= 0.3 ;
    env.epsilon = 0.01 ;
    env.delta   = 0. ;
    env.pacing  = false ;
    env.pacePeriod = 200 ;

    var SolverUniforms = function(_TRGT){
        this.inTrgt     = { type : 's', value : _TRGT           } ;
        this.dt         = { type : 'f', value : env.dt          } ;
        this.aa         = { type : 'f', value : env.aa          } ;
        this.bb         = { type : 'f', value : env.bb          } ;
        this.epsilon    = { type : 'f', value : env.epsilon     } ;
        this.delta      = { type : 'f', value : env.delta       } ;
        this.pacePeriod = { type : 'f', value : env.pacePeriod  } ;
        return this ;
    }

    /* Setup a solver for odd time steps */
    env.fsol = new Abubu.Solver({
        fragmentShader  : compShader.value ,
        vertexShader    : vertShader.value ,
        uniforms : new SolverUniforms(env.ftrgt) ,
        renderTargets: { 
            outTrgt : { location : 0 , target : env.strgt } ,
        }
    } ) ;
    /* Setup a solver for an even step */
    env.ssol = new Abubu.Solver({
        fragmentShader  : compShader.value ,
        vertexShader    : vertShader.value ,
        uniforms : new SolverUniforms(env.strgt),
        renderTargets: { 
            outTrgt : { location : 0 , target : env.ftrgt } ,
        }
    } ) ;

    env.plot = new Abubu.Plot2D({
            target : env.strgt,
            channel: 'r' ,
            minValue : 0. ,
            maxValue : 1. ,
            colorbar : true ,
            canvas : canvas_1, 
    } ) ;


    env.initialize = function(){
        env.init.render() ;
        env.time = 0 ;
    }
    env.plot.init() ;
    env.running = false ;

    env.solve = function(){
        env.running = !env.running ;
    }

    env.initialize() ;
    createGUI() ;
    render() ;
}/*  End of loadWebGL  */

function render(){
    if(env.running){
        for(var i=0; i<40 ; i++){
            env.time += env.dt*2. ;
            env.fsol.render() ;
            env.ssol.render() ;
        }
    }
    env.plot.render() ;
    requestAnimationFrame(render) ;
}

function createGUI(){
    env.gui = new Abubu.Gui() ;
    env.pnl = env.gui.addPanel() ;
    var pnl = env.pnl ;
    pnl.add(env, 'initialize') ;
    pnl.add(env, 'solve').name('Sove/Stop') ;

    pnl.add(env, 'aa').onChange(function(){
        env.fsol.uniforms.aa.value = env.aa ;
        env.ssol.uniforms.aa.value = env.aa ;
    }) ;
    pnl.add(env, 'bb').onChange(function(){
        env.fsol.uniforms.bb.value = env.bb ;
        env.ssol.uniforms.bb.value = env.bb ;
    }) ;
    pnl.add(env, 'epsilon').onChange(function(){
        env.fsol.uniforms.epsilon.value = env.epsilon ;
        env.ssol.uniforms.epsilon.value = env.epsilon ;
    }) ;
    pnl.add(env, 'delta').onChange(function(){
        env.fsol.uniforms.delta.value = env.delta ;
        env.ssol.uniforms.delta.value = env.delta ;
    }).step(0.01) ;
    
    pnl.add(env, 'pacePeriod').onChange(function(){
        env.fsol.uniforms.pacePeriod.value = env.pacePeriod ;
        env.ssol.uniforms.pacePeriod.value = env.pacePeriod ;
    }).step(0.01) ;
}

loadWebGL() ;

} ) ;   // End of require 
